# -*- coding: utf-8 -*-
from django.utils.translation import ugettext_lazy as _
from django.utils.html import format_html
from django.template.loader import select_template
from django.forms.fields import ChoiceField
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.link.forms import LinkForm
from shop import app_settings
from shop.cascade.plugin_base import DialogFormPluginBase, ShopLinkPluginBase, ShopLinkElementMixin
from shop.cascade.checkout import CustomerFormPluginBase
from .forms import SubscribeForm, ConfirmForm_factory


class CustomerFormMinimalPlugin(CustomerFormPluginBase):
    """A modified copy of CustomerFormPlugin"""
    name = _("Customer Form (minimal)")
    form_class = 'shop_subscribe.forms.CustomerFormMinimal'
    def render(self, context, instance, placeholder):
        if not context['request'].customer.is_registered():
            context['error_message'] = _("Only registered customers can access this form.")
            return context
        return super(CustomerFormMinimalPlugin, self).render(context, instance, placeholder)

DialogFormPluginBase.register_plugin(CustomerFormMinimalPlugin)


SUBSCRIPTION_FORM_TYPES = (
    ('subscribe', _("Subscribe Form"), SubscribeForm),
    ('confirm', _("Email Confirmation Form"), ConfirmForm_factory()),
)

class SubscriptionsAdminForm(LinkForm):
    LINK_TYPE_CHOICES = (('DO_NOTHING', _("Do Nothing")), ('RELOAD_PAGE', _("Reload Page")), ('cmspage', _("CMS Page")))
    form_type = ChoiceField(label=_("Rendered Form"), choices=(ft[:2] for ft in SUBSCRIPTION_FORM_TYPES),
        help_text=_("Select the required form."))

    def clean(self):
        cleaned_data = super(SubscriptionsAdminForm, self).clean()
        if self.is_valid():
            cleaned_data['glossary'].update(form_type=cleaned_data['form_type'])
        return cleaned_data

class SubscriptionsFormPlugin(ShopLinkPluginBase):
    """
    Cannot use the DialogFormPlugin framework since the form requires a cart which may not exist
    Use the shop GuestForm plugin for the checkout process
    """
    name = _("Subscriptions Form")
    require_parent = True
    parent_classes = ('BootstrapColumnPlugin', 'BootstrapPanelPlugin', 'SegmentPlugin', 'SimpleWrapperPlugin')
    model_mixins = (ShopLinkElementMixin,)
    cache = False
    form = SubscriptionsAdminForm
    fields = ('form_type', ('link_type', 'cms_page'), 'glossary',)

    @classmethod
    def get_identifier(cls, instance):
        identifier = super(SubscriptionsFormPlugin, cls).get_identifier(instance)
        content = dict(ft[:2] for ft in SUBSCRIPTION_FORM_TYPES).get(instance.glossary.get('form_type'), _("unknown"))
        return format_html('{0}{1}', identifier, content)

    def get_render_template(self, context, instance, placeholder):
        form_type = instance.glossary.get('form_type')
        template_names = [
            '{}/shop_subscribe/{}-form.html'.format(app_settings.APP_LABEL, form_type),
            'shop_subscribe/{}-form.html'.format(form_type),
        ]
        return select_template(template_names)

    def render(self, context, instance, placeholder):
        """
        Return the context to render the template
        """
        form_type = instance.glossary.get('form_type')
        form_class = dict( ((ft[0], ft[2]) for ft in SUBSCRIPTION_FORM_TYPES) ).get(form_type, None)
        context['form'] = form_class(request=context['request'])
        context['action'] = self.get_link(instance)
        return super(SubscriptionsFormPlugin, self).render(context, instance, placeholder)

plugin_pool.register_plugin(SubscriptionsFormPlugin)



# -*- coding: utf-8 -*-
from rest_framework import serializers
from shop.models.customer import CustomerModel
from .utils import get_subscription_fields


# These serializers are for debugging only via the DRF browsable api and are not required for the form operations
class SubscribeSerializer(serializers.ModelSerializer):
    """Allows email address submissions"""
    email = serializers.EmailField()
    class Meta:
        model = CustomerModel
        fields = ('email',)


def ConfirmSerializer_factory():
    """Dynamically generate fields list"""

    subscription_fields = get_subscription_fields()

    class ConfirmSerializer(serializers.ModelSerializer):
        email = serializers.EmailField()
        sig = serializers.CharField(min_length=20)

        class Meta:
            model = CustomerModel
            fields = ['email', 'sig'] + subscription_fields

    return ConfirmSerializer



# -*- coding: utf-8 -*-
from django.conf.urls import url
from .views import SubscribeView, ConfirmView


app_name = 'shop_subscribe'
urlpatterns = [
    url(r'^subscribe/$', SubscribeView.as_view(), name='subscribe'),
    url(r'^confirm/$', ConfirmView.as_view(), name='confirm'),
]



# -*- coding: utf-8 -*-
from django.forms import fields, widgets, ValidationError
from django.urls import reverse
from django.core.signing import BadSignature
from django.utils.translation import ugettext_lazy as _
from djng.forms import NgModelFormMixin, NgFormValidationMixin
from djng.forms.angular_base import SafeTuple
from djng.styling.bootstrap3.forms import Bootstrap3ModelForm
from shop.forms.checkout import CustomerForm
from shop.models.customer import CustomerModel
from .utils import send_confirmation_email, logger, unsign, get_customer_from_emailsignature, get_subscription_fields


class CustomerFormMinimal(CustomerForm):
    """For use instead of the shop CustomerForm Plugin in checkout without the other customer fields like subscriptions"""
    form_name = 'customer_form'  # Override form name to reuse template `customer-form.html`
    class Meta:
        model = CustomerModel
        custom_fields = ('email', 'first_name', 'last_name')
        fields = ('plugin_id', 'plugin_order', 'email', 'first_name', 'last_name')


def debug_instance(instance):
    """Helper debug function"""
    print("Registration form instance: '{}', ".format(instance), end='')
    if instance:
        try:
            if instance.is_visitor():
                print('Visiting Customer')
            elif instance.is_authenticated():
                print('Registered Customer')
            elif instance.is_guest():
                print('Guest Customer')
            elif instance.is_anonymous():
                print('Unrecognized Customer')
            else:
                print('Unknown Concrete Customer')
        except AttributeError:
            print('User')
    else:
        print('None')


class NgSuccessMessageMixin(object):
    success_message = _("Form submitted successfully")
    def non_field_errors(self):
        errors = super(NgSuccessMessageMixin, self).non_field_errors()
        errors.append(SafeTuple((self.form_name, self.form_error_css_classes, '$dirty', 'success', 'valid', self.success_message)))
        return errors


class SubscribeForm(NgSuccessMessageMixin, NgModelFormMixin, NgFormValidationMixin, Bootstrap3ModelForm):
    email = fields.EmailField(label=_('Subscribe to product updates'), required=True)

    form_name = 'subscribe_form'
    scope_prefix = 'subscribe_data'
    required_css_class = 'djng-field-required'
    success_message = _("Thanks for subscribing!")

    class Meta:
        model = CustomerModel
        fields = ('email',)

    @classmethod
    def get_url(cls):
        """Returns the url for the form submission"""
        return reverse('shop_subscribe:subscribe')

    def __init__(self, request=None, instance=None, *args, **kwargs):
        """
        Always called with a Customer instance, not User.
        """
        #debug_instance(request.customer)
        if request is None:
            raise ValueError("Request must not be None")
        if instance:
            raise ValueError("Pass in 'request' instead of 'instance'")

        # for save()
        self.request = request

        if request.customer.is_visitor():
            request.customer = CustomerModel.objects.get_or_create_from_request(request)
        super(SubscribeForm, self).__init__(instance=request.customer, *args, **kwargs)

    # custom
    unchecked_error = _("At least one subscription must be checked")
    def clean_subscriptions(self):
        """
        Signups must have at least one subscription checked.
        Use this in your clean function if overriding this form.
        """
        cleaned_data = super(SignupSubscriptionForm, self).clean() or self.cleaned_data
        checked = 0
        for key, value in cleaned_data.items():
            if key.startswith('subscription') and value:
                checked += 1
        if checked > 0:
            return cleaned_data
        else:
            raise ValidationError(self.unchecked_error)

    registered_error = _("The current user is already registered")
    def clean_email(self):
        """
        Reject requests for sessions with email addresses
        Do not lookup user with current email and assign to customer as it is privilege escalation
        Also don't provide ValidationError that email exists as can be used to check db contents
        """
        if getattr(self.instance, 'email', None):
            raise ValidationError(self.registered_error)
        return self.cleaned_data['email']

    def save(self, **kwargs):
        """
        Only save if email address doesn't already exist in the db
        A confirmation email address will be sent where customers can change subscriptions
        """
        if CustomerModel.objects.filter(user__email=self.cleaned_data['email']).exists():
            logger.info('Subscription from {} dropped, email address exists.'.format(self.cleaned_data['email']))
            return self.instance
        # email is not assigned by the form probably because it is a related user object field
        self.instance.email = self.cleaned_data['email']
        if send_confirmation_email(self.request, self.instance):
            self.instance = super(SubscribeForm, self).save(**kwargs)
        return self.instance


class NgLoadSuccessMessageMixin(object):
    loadsuccess_message = _("Form loaded successfully")
    def non_field_errors(self):
        errors = super(NgLoadSuccessMessageMixin, self).non_field_errors()
        errors.append(SafeTuple((self.form_name, self.form_error_css_classes, '$pristine', 'email.$modelValue', 'valid', self.loadsuccess_message)))
        return errors


def ConfirmForm_factory():
    """Dynamically generate fields list"""

    subscription_fields = get_subscription_fields()

    class ConfirmForm(NgLoadSuccessMessageMixin, NgSuccessMessageMixin, NgModelFormMixin, NgFormValidationMixin, Bootstrap3ModelForm):
        "email and sig are obtained from a url and embedded in the form data hidden for posts"
        email = fields.EmailField(required=True, widget=widgets.HiddenInput)
        sig = fields.CharField(min_length=20, required=True, widget=widgets.HiddenInput)

        form_name = 'confirm_form'
        scope_prefix = 'confirm_data'
        required_css_class = 'djng-field-required'
        success_message = _("Subscriptions updated")
        loadsuccess_message = _("Email address confirmed")

        class Meta:
            model = CustomerModel
            fields = ['email', 'sig'] + subscription_fields

        @classmethod
        def get_url(cls):
            """Returns the url for the form submission"""
            return reverse('shop_subscribe:confirm')

        def __init__(self, request=None, instance=None, *args, **kwargs):
            """
            Always called with a Customer instance, not User.
            """
            if request is None:
                raise ValueError("Request must not be None")
            if instance:
                raise ValueError("Pass in 'request' instead of 'instance'")

            try:
                customer, initial = get_customer_from_emailsignature(request)
            except BadSignature:
                # if no/bad url params are given, uses a blank form
                customer, initial = None, None

            super(ConfirmForm, self).__init__(instance=customer, initial=initial, *args, **kwargs)

        confirm_error = _("Could not confirm the supplied email address.")
        def clean(self):
            """
            Checks the email signature.
            is_valid or full_clean won't check initial values unless data has been passed in
            """
            cleaned_data = super(ConfirmForm, self).clean() or self.cleaned_data
            try:
                unsign(cleaned_data)
                return cleaned_data
            except BadSignature:
                raise ValidationError(self.confirm_error)

    return ConfirmForm



# -*- coding: utf-8 -*-
from django.utils.translation import ugettext_lazy as _
from shop.admin.customer import CustomerInlineAdminBase
from .utils import get_subscription_fields


class SubscriptionsInlineAdmin(CustomerInlineAdminBase):
    """Add to CustomerInline for our fields and to take advantage of JS inline move to top"""
    fieldsets = list(CustomerInlineAdminBase.fieldsets) + [
        (_("Subscriptions"),        {'fields': get_subscription_fields()})
    ]



# -*- coding: utf-8 -*-
from collections import OrderedDict
from datetime import datetime
import re, logging
from django.core.signing import Signer, BadSignature
from django.template.loader import select_template
from django.contrib.sites.shortcuts import get_current_site
from django.utils.translation import get_language_from_request
from django.utils import timezone
from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist
from django.urls import NoReverseMatch
from cms.cache.page import get_page_url_cache, set_page_url_cache
from cms.templatetags.cms_tags import _get_page_by_untyped_arg
#from cms.utils import get_language_from_request
from rest_framework.request import Request as DRFRequest
from bs4 import BeautifulSoup, NavigableString, Tag
from ipware.ip import get_real_ip, get_ip
from shop import app_settings
from shop.models.customer import CustomerModel


# Get an instance of a logger
logger = logging.getLogger('shop_subscribe')

sep = ':'
signer = Signer(sep=sep, salt='subscribe')
def sign(email):
    "Returns [email, signature]"
    return signer.sign(email).split(sep, 1)

def unsign(context):
    "Returns an ordereddict of email & signature, {} otherwise"
    sigcontext = OrderedDict([
        ('email', context.get('email', '')),
        ('sig', context.get('sig', ''))
    ])
    signer.unsign(sep.join( sigcontext.values() ))
    return sigcontext


def html_to_text(html):
    "Creates a formatted text email message as a string from a rendered html template (page)"
    soup = BeautifulSoup(html, 'html.parser')
    # Ignore anything in head
    body, text = soup.body, []
    for element in body.descendants:
        # We use type and not isinstance since comments, cdata, etc are subclasses that we don't want
        if type(element) == NavigableString:
            parent_tags = (t for t in element.parents if type(t) == Tag)
            hidden = False
            for parent_tag in parent_tags:
                # Ignore any text inside a non-displayed tag
                # We also behave is if scripting is enabled (noscript is ignored)
                # The list of non-displayed tags and attributes from the W3C specs:
                if (parent_tag.name in ('area', 'base', 'basefont', 'datalist', 'head', 'link',
                                        'meta', 'noembed', 'noframes', 'param', 'rp', 'script',
                                        'source', 'style', 'template', 'track', 'title', 'noscript') or
                    parent_tag.has_attr('hidden') or
                    (parent_tag.name == 'input' and parent_tag.get('type') == 'hidden')):
                    hidden = True
                    break
            if hidden:
                continue

            # remove any multiple and leading/trailing whitespace
            string = ' '.join(element.string.split())
            if string:
                if element.parent.name == 'a':
                    a_tag = element.parent
                    # replace link text with the link
                    string = a_tag['href']
                    # concatenate with any non-empty immediately previous string
                    if (    type(a_tag.previous_sibling) == NavigableString and
                            a_tag.previous_sibling.string.strip() ):
                        text[-1] = text[-1] + ' ' + string
                        continue
                elif element.previous_sibling and element.previous_sibling.name == 'a':
                    text[-1] = text[-1] + ' ' + string
                    continue
                elif element.parent.name == 'p':
                    # Add extra paragraph formatting newline
                    string = '\n' + string
                text += [string]
    doc = '\n'.join(text)
    return doc


def get_subscription_fields():
    """Returns the list of customer subscription fields"""
    return [item for item in dir(CustomerModel) if item.startswith('subscription_')]


def get_customer_from_emailsignature(request):
    """
    DRF or WSGI requests
    Validate the email signature in either the GET url for initial email link or POST hidden data for form submissions.
    If the signature is valid return a 'recognized' customer object if not already.
    """
    if isinstance(request, DRFRequest):
        try:
            # e.g. GET URLs
            context = unsign(request.query_params)
        except BadSignature:
            # e.g. POST data
            context = unsign(request.data)
    else:
        try:
            context = unsign(request.GET)
        except BadSignature:
            context = unsign(request.POST)

    customer, created = CustomerModel.objects.get_or_create_from_email(request, context['email'])
    # make the customer guest if not already
    if not customer.is_recognized():
        customer.recognize_as_guest()
    # remove sensitive data since the customer has confirmed
    customer.extra.pop('subscription_IP', None)
    customer.extra.pop('subscription_date', None)
    customer.last_access = timezone.now()
    customer.save()

    if created:
        logger.warning('Customer recreated from confirmation: %s' % context['email'])

    return customer, context


def reverse_cms_url(request, page_lookup):
    """Equivalent of CMS 'page_url' templatetag"""
    site_id = get_current_site(request).id
    lang = get_language_from_request(request)

    url = get_page_url_cache(page_lookup, lang, site_id)
    if url is None:
        try:
            page = _get_page_by_untyped_arg(page_lookup, request, site_id)
        except ObjectDoesNotExist:
            page = None
        if page:
            url = page.get_absolute_url(language=lang)
            set_page_url_cache(page_lookup, lang, site_id, url)
    if url:
        return url
    raise NoReverseMatch("CMS page not found")

def build_confirm_url(request, email='', sig=''):
    """
    Build the confirm url from supplied parameters
    """
    try:
        url = reverse_cms_url(request, 'shop-subscribe-confirm')
    except NoReverseMatch:
        try:
            url = reverse('shop-subscribe-confirm')
        except NoReverseMatch:
            url = reverse('shop_subscribe:confirm')
    url = url + '?email=' + '&sig='.join( (email, sig) )
    return request.build_absolute_uri(url)

def unisoformat(datestring):
    "Bit of a hacky way to get the date back from python's isoformat string"
    return datetime(*map(int, re.findall('\d+', datestring)))

def send_confirmation_email(request, customer):
    """
    Uses settings EMAIL_BACKEND to send emails
    Assumes customer will be saved afterward externally
    """
    # check that same IP is not making lots of subscriptions and store IP
    ip = get_ip(request)
    now = datetime.now()
    for c in CustomerModel.objects.exclude(extra__exact='{}').all():
        try:
            if (c.extra['subscription_IP'] == ip and
                (now - unisoformat(c.extra['subscription_date'])).days < 1):
                logger.warning('Subscription from {} dropped. Same IP ({}) subscribed recently.'.format(customer.email, ip))
                return False
        except KeyError:
            continue

    context = {
        'site_name': get_current_site(request).name,
        'confirm_url': build_confirm_url(request, *sign(customer.email)),
        'email': customer.email,
        # requires django-ipware; only returns public IPs
        'ip': get_real_ip(request),
        'user_agent': request.META['HTTP_USER_AGENT'],
        'language': get_language_from_request(request, check_path=True)
    }
    subject = select_template([
        '{}/shop_subscribe/email/subscription-confirm-subject.txt'.format(app_settings.APP_LABEL),
        'shop_subscribe/email/subscription-confirm-subject.txt',
    ]).render(context)
    # Email subject *must not* contain newlines
    subject = ' '.join(subject.splitlines()).strip()
    html = select_template([
        '{}/shop_subscribe/email/subscription-confirm-body.html'.format(app_settings.APP_LABEL),
        'shop_subscribe/email/subscription-confirm-body.html',
    ]).render(context)
    text = html_to_text(html)
    customer.user.email_user(subject, text, html_message=html)

    customer.extra.update({
        'subscription_IP': ip,
        'subscription_date': now.isoformat(),
    })
    #customer.save(update_fields=["extra"])
    return True
